@@ -1,7 +1,9 @@
 #!/usr/bin/env python
 
+import base64
 import datetime
 import hashlib
+import logging
 import math
 import operator
 import optparse
@@ -15,6 +17,8 @@ from collections import namedtuple, OrderedDict
 from functools import wraps
 from getpass import getpass
 from io import TextIOWrapper
+from logging.handlers import WatchedFileHandler
+from werkzeug.routing import BaseConverter
 
 # Py2k compat.
 if sys.version_info[0] == 2:
@@ -30,15 +34,21 @@ else:
     decode_handler = 'backslashreplace'
     numeric = (int, float)
     unicode_type = str
+    from functools import reduce
     from io import StringIO
 
 try:
     from flask import (
-        Flask, abort, escape, flash, jsonify, make_response, Markup, redirect,
+        Flask, abort, flash, jsonify, make_response, redirect,
         render_template, request, session, url_for)
 except ImportError:
     raise RuntimeError('Unable to import flask module. Install by running '
                        'pip install flask')
+try:
+    from markupsafe import Markup, escape
+except ImportError:
+    raise RuntimeError('Unable to import markupsafe module. Install by running'
+                       ' pip install markupsafe')
 
 try:
     from pygments import formatters, highlight, lexers
@@ -162,6 +172,26 @@ class SqliteDataSet(DataSet):
             ('trigger',))
         return [TriggerMetadata(*row) for row in cursor.fetchall()]
 
+    def get_table_sql(self, table):
+        if not table:
+            return
+
+        cursor = dataset.query(
+            'SELECT sql FROM sqlite_master '
+            'WHERE tbl_name = ? AND type IN (?, ?)',
+            [table, 'table', 'view'])
+        res = cursor.fetchone()
+        if res is not None:
+            return res[0]
+
+    def get_view(self, name):
+        cursor = self.query(
+            'SELECT name, sql FROM sqlite_master '
+            'WHERE type = ? AND name = ?', ('view', name))
+        res = cursor.fetchone()
+        if res is not None:
+            return ViewMetadata(*res)
+
     def get_all_views(self):
         cursor = self.query(
             'SELECT name, sql FROM sqlite_master '
@@ -184,6 +214,38 @@ class SqliteDataSet(DataSet):
             '%s_%s' % (virtual_table, suffix) for suffix in suffixes
             for virtual_table in virtual_tables)
 
+    def is_view(self, name):
+        cursor = self.query(
+            'SELECT name FROM sqlite_master '
+            'WHERE type = ? AND name = ?', ('view', name))
+        return cursor.fetchone() is not None
+
+    def view_operations(self, name):
+        cursor = self.query(
+            'SELECT sql FROM sqlite_master WHERE type=? AND tbl_name=?',
+            ('trigger', name))
+        triggers = [t for t, in cursor.fetchall()]
+        rgx = re.compile(r'CREATE\s+TRIGGER.+?\sINSTEAD\s+OF\s+'
+                         '(INSERT|UPDATE|DELETE)\s', re.I)
+        operations = set()
+        for trigger in triggers:
+            operations.update([op.lower() for op in rgx.findall(trigger)])
+
+        return operations
+
+class Base64Converter(BaseConverter):
+    def to_python(self, value):
+        return base64.urlsafe_b64decode(value).decode()
+
+    def to_url(self, value):
+        if not isinstance(value, (bytes, str)):
+            value = str(value).encode('utf8')
+        elif isinstance(value, str):
+            value = value.encode('utf8')
+        return base64.urlsafe_b64encode(value).decode()
+
+app.url_map.converters['b64'] = Base64Converter
+
 #
 # Flask views.
 #
@@ -199,6 +261,8 @@ def login():
             session['authorized'] = True
             return redirect(session.get('next_url') or url_for('index'))
         flash('The password you entered is incorrect.', 'danger')
+        app.logger.debug('Received incorrect password attempt from %s' %
+                         request.remote_addr)
     return render_template('login.html')
 
 @app.route('/logout/', methods=['GET'])
@@ -206,6 +270,70 @@ def logout():
     session.pop('authorized', None)
     return redirect(url_for('login'))
 
+def _query_view(template, table=None):
+    data = []
+    data_description = error = row_count = sql = None
+    ordering = None
+
+    sql = qsql = request.values.get('sql') or ''
+
+    if 'export_json' in request.values:
+        ordering = request.values.get('export_ordering')
+        export_format = 'json'
+    elif 'export_csv' in request.values:
+        ordering = request.values.get('export_ordering')
+        export_format = 'csv'
+    else:
+        ordering = request.values.get('ordering')
+        export_format = None
+
+    if ordering:
+        ordering = int(ordering)
+        direction = 'DESC' if ordering < 0 else 'ASC'
+        qsql = ('SELECT * FROM (%s) AS _ ORDER BY %d %s' %
+                (sql.rstrip(' ;'), abs(ordering), direction))
+    else:
+        ordering = None
+
+    if table:
+        default_sql = 'SELECT * FROM "%s"' % table
+        model_class = dataset[table].model_class
+    else:
+        default_sql = ''
+        model_class = dataset._base_model
+
+    if qsql:
+        if export_format:
+            query = model_class.raw(qsql).dicts()
+            return export(query, export_format, table)
+
+        try:
+            cursor = dataset.query(qsql)
+        except Exception as exc:
+            error = str(exc)
+            app.logger.exception('Error in user-submitted query.')
+        else:
+            data = cursor.fetchall()[:app.config['MAX_RESULT_SIZE']]
+            data_description = cursor.description
+            row_count = cursor.rowcount
+
+    return render_template(
+        template,
+        data=data,
+        data_description=data_description,
+        default_sql=default_sql,
+        error=error,
+        ordering=ordering,
+        query_images=get_query_images(),
+        row_count=row_count,
+        sql=sql,
+        table=table,
+        table_sql=dataset.get_table_sql(table))
+
+@app.route('/query/', methods=['GET', 'POST'])
+def generic_query():
+    return _query_view('query.html')
+
 def require_table(fn):
     @wraps(fn)
     def inner(table, *args, **kwargs):
@@ -224,7 +352,11 @@ def table_create():
             dest = '/' + dest
         return redirect(dest)
 
-    dataset[table]
+    try:
+        dataset[table]
+    except Exception as exc:
+        flash('Error: %s' % str(exc), 'danger')
+        app.logger.exception('Error attempting to create table.')
     return redirect(url_for('table_import', table=table))
 
 @app.route('/<table>/')
@@ -233,10 +365,6 @@ def table_structure(table):
     ds_table = dataset[table]
     model_class = ds_table.model_class
 
-    table_sql = dataset.query(
-        'SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = ?',
-        [table, 'table']).fetchone()[0]
-
     return render_template(
         'table_structure.html',
         columns=dataset.get_columns(table),
@@ -245,7 +373,7 @@ def table_structure(table):
         indexes=dataset.get_indexes(table),
         model_class=model_class,
         table=table,
-        table_sql=table_sql,
+        table_sql=dataset.get_table_sql(table),
         triggers=dataset.get_triggers(table))
 
 def get_request_data():
@@ -256,17 +384,20 @@ def get_request_data():
 @app.route('/<table>/add-column/', methods=['GET', 'POST'])
 @require_table
 def add_column(table):
+    class JsonField(TextField):
+        field_type = 'JSON'
     column_mapping = OrderedDict((
-        ('VARCHAR', CharField),
         ('TEXT', TextField),
         ('INTEGER', IntegerField),
         ('REAL', FloatField),
-        ('BOOL', BooleanField),
         ('BLOB', BlobField),
+        ('JSON', JsonField),
+        ('BOOL', BooleanField),
         ('DATETIME', DateTimeField),
         ('DATE', DateField),
+        ('DECIMAL', DecimalField),
         ('TIME', TimeField),
-        ('DECIMAL', DecimalField)))
+        ('VARCHAR', CharField)))
 
     request_data = get_request_data()
     col_type = request_data.get('type')
@@ -274,28 +405,30 @@ def add_column(table):
 
     if request.method == 'POST':
         if name and col_type in column_mapping:
-            migrate(
-                migrator.add_column(
-                    table,
-                    name,
-                    column_mapping[col_type](null=True)))
-            flash('Column "%s" was added successfully!' % name, 'success')
-            dataset.update_cache(table)
-            return redirect(url_for('table_structure', table=table))
+            try:
+                migrate(
+                    migrator.add_column(
+                        table,
+                        name,
+                        column_mapping[col_type](null=True)))
+            except Exception as exc:
+                flash('Error attempting to add column "%s": %s' % (name, exc),
+                      'danger')
+                app.logger.exception('Error attempting to add column.')
+            else:
+                flash('Column "%s" was added successfully!' % name, 'success')
+                dataset.update_cache(table)
+                return redirect(url_for('table_structure', table=table))
         else:
             flash('Name and column type are required.', 'danger')
 
-    table_sql = dataset.query(
-        'SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = ?',
-        [table, 'table']).fetchone()[0]
-
     return render_template(
         'add_column.html',
         col_type=col_type,
         column_mapping=column_mapping,
         name=name,
         table=table,
-        table_sql=table_sql)
+        table_sql=dataset.get_table_sql(table))
 
 @app.route('/<table>/drop-column/', methods=['GET', 'POST'])
 @require_table
@@ -307,10 +440,16 @@ def drop_column(table):
 
     if request.method == 'POST':
         if name in column_names:
-            migrate(migrator.drop_column(table, name))
-            flash('Column "%s" was dropped successfully!' % name, 'success')
-            dataset.update_cache(table)
-            return redirect(url_for('table_structure', table=table))
+            try:
+                migrate(migrator.drop_column(table, name))
+            except Exception as exc:
+                flash('Error attempting to drop column "%s": %s' % (name, exc),
+                      'danger')
+                app.logger.exception('Error attempting to drop column.')
+            else:
+                flash('Column "%s" was dropped successfully!' % name, 'success')
+                dataset.update_cache(table)
+                return redirect(url_for('table_structure', table=table))
         else:
             flash('Name is required.', 'danger')
 
@@ -333,10 +472,16 @@ def rename_column(table):
 
     if request.method == 'POST':
         if (rename in column_names) and (rename_to not in column_names):
-            migrate(migrator.rename_column(table, rename, rename_to))
-            flash('Column "%s" was renamed successfully!' % rename, 'success')
-            dataset.update_cache(table)
-            return redirect(url_for('table_structure', table=table))
+            try:
+                migrate(migrator.rename_column(table, rename, rename_to))
+            except Exception as exc:
+                flash('Error attempting to rename column "%s": %s' % (name, exc),
+                      'danger')
+                app.logger.exception('Error attempting to rename column.')
+            else:
+                flash('Column "%s" was renamed successfully!' % rename, 'success')
+                dataset.update_cache(table)
+                return redirect(url_for('table_structure', table=table))
         else:
             flash('Column name is required and cannot conflict with an '
                   'existing column\'s name.', 'danger')
@@ -360,13 +505,18 @@ def add_index(table):
 
     if request.method == 'POST':
         if indexed_columns:
-            migrate(
-                migrator.add_index(
-                    table,
-                    indexed_columns,
-                    unique))
-            flash('Index created successfully.', 'success')
-            return redirect(url_for('table_structure', table=table))
+            try:
+                migrate(
+                    migrator.add_index(
+                        table,
+                        indexed_columns,
+                        unique))
+            except Exception as exc:
+                flash('Error attempting to create index: %s' % exc, 'danger')
+                app.logger.exception('Error attempting to create index.')
+            else:
+                flash('Index created successfully.', 'success')
+                return redirect(url_for('table_structure', table=table))
         else:
             flash('One or more columns must be selected.', 'danger')
 
@@ -387,9 +537,14 @@ def drop_index(table):
 
     if request.method == 'POST':
         if name in index_names:
-            migrate(migrator.drop_index(table, name))
-            flash('Index "%s" was dropped successfully!' % name, 'success')
-            return redirect(url_for('table_structure', table=table))
+            try:
+                migrate(migrator.drop_index(table, name))
+            except Exception as exc:
+                flash('Error attempting to drop index: %s' % exc, 'danger')
+                app.logger.exception('Error attempting to drop index.')
+            else:
+                flash('Index "%s" was dropped successfully!' % name, 'success')
+                return redirect(url_for('table_structure', table=table))
         else:
             flash('Index name is required.', 'danger')
 
@@ -410,9 +565,14 @@ def drop_trigger(table):
 
     if request.method == 'POST':
         if name in trigger_names:
-            dataset.query('DROP TRIGGER "%s";' % name)
-            flash('Trigger "%s" was dropped successfully!' % name, 'success')
-            return redirect(url_for('table_structure', table=table))
+            try:
+                dataset.query('DROP TRIGGER "%s";' % name)
+            except Exception as exc:
+                flash('Error attempting to drop trigger: %s' % exc, 'danger')
+                app.logger.exception('Error attempting to drop trigger.')
+            else:
+                flash('Trigger "%s" was dropped successfully!' % name, 'success')
+                return redirect(url_for('table_structure', table=table))
         else:
             flash('Trigger name is required.', 'danger')
 
@@ -423,17 +583,21 @@ def drop_trigger(table):
         name=name,
         table=table)
 
+
 @app.route('/<table>/content/')
 @require_table
 def table_content(table):
     page_number = request.args.get('page') or ''
+    if page_number == 'last': page_number = '1000000'
     page_number = int(page_number) if page_number.isdigit() else 1
 
     dataset.update_cache(table)
     ds_table = dataset[table]
+    model = ds_table.model_class
+
     total_rows = ds_table.all().count()
     rows_per_page = app.config['ROWS_PER_PAGE']
-    total_pages = int(math.ceil(total_rows / float(rows_per_page)))
+    total_pages = max(1, int(math.ceil(total_rows / float(rows_per_page))))
     # Restrict bounds.
     page_number = min(page_number, total_pages)
     page_number = max(page_number, 1)
@@ -445,18 +609,16 @@ def table_content(table):
 
     ordering = request.args.get('ordering')
     if ordering:
-        field = ds_table.model_class._meta.columns[ordering.lstrip('-')]
+        field = model._meta.columns[ordering.lstrip('-')]
         if ordering.startswith('-'):
             field = field.desc()
         query = query.order_by(field)
 
+    session['%s.last_viewed' % table] = (page_number, ordering)
+
     field_names = ds_table.columns
     columns = [f.column_name for f in ds_table.model_class._meta.sorted_fields]
 
-    table_sql = dataset.query(
-        'SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = ?',
-        [table, 'table']).fetchone()[0]
-
     return render_template(
         'table_content.html',
         columns=columns,
@@ -468,86 +630,297 @@ def table_content(table):
         previous_page=previous_page,
         query=query,
         table=table,
+        table_pk=model._meta.primary_key,
+        table_sql=dataset.get_table_sql(table),
         total_pages=total_pages,
         total_rows=total_rows)
 
+def minimal_validate_field(field, value):
+    if value.lower().strip() == 'null':
+        value = None
+    if value is None and not field.null:
+        return 'NULL', 'Column does not allow NULL values.'
+    if value is None:
+        return None, None
+    if isinstance(field, IntegerField) and not value.isdigit():
+        return value, 'Value is not a number.'
+    elif isinstance(field, FloatField):
+        try:
+            _ = float(value)
+        except Exception:
+            return value, 'Value is not a numeric/real.'
+    elif isinstance(field, BooleanField):
+        if value.lower() not in ('1', '0', 'true', 'false', 't', 'f'):
+            return value, 'Value must be 1, 0, true, false, t or f.'
+        value = True if value.lower() in ('1', 't', 'true') else False
+    elif isinstance(field, BlobField):
+        try:
+            value = base64.b64decode(value)
+        except Exception as exc:
+            return value, 'Value must be base64-encoded binary data.'
+    try:
+        field.db_value(value)
+    except Exception as exc:
+        return value, str(exc)
+
+    return value, None
+
+@app.route('/<table>/insert/', methods=['GET', 'POST'])
+@require_table
+def table_insert(table):
+    dataset.update_cache(table)
+    model = dataset[table].model_class
+
+    columns = []
+    col_dict = {}
+    row = {}
+    for column in dataset.get_columns(table):
+        field = model._meta.columns[column.name]
+        if isinstance(field, AutoField):
+            continue
+        columns.append(column)
+        col_dict[column.name] = column
+        row[column.name] = ''
+
+    edited = set()
+    errors = {}
+    if request.method == 'POST':
+        insert = {}
+        for key, value in request.form.items():
+            if key not in col_dict: continue
+            column = col_dict[key]
+            edited.add(column.name)
+            row[column.name] = value
+
+            field = model._meta.columns[column.name]
+            value, err = minimal_validate_field(field, value)
+            if err:
+                errors[key] = err
+            else:
+                insert[field] = value
+
+        if errors:
+            flash('One or more errors prevented the row being inserted.',
+                  'danger')
+        elif insert:
+            try:
+                with dataset.transaction() as txn:
+                    n = model.insert(insert).execute()
+            except Exception as exc:
+                flash('Insert failed: %s' % exc, 'danger')
+                app.logger.exception('Error attempting to insert row into %s.', table)
+            else:
+                flash('Successfully inserted record (%s).' % n, 'success')
+                return redirect(url_for(
+                    'table_content',
+                    table=table,
+                    page='last'))
+        else:
+            flash('No data was specified to be inserted.', 'warning')
+    else:
+        edited = set(col_dict)  # Make all fields editable on load.
+
+    return render_template(
+        'table_insert.html',
+        columns=columns,
+        edited=edited,
+        errors=errors,
+        model=model,
+        row=row,
+        table=table)
+
+def redirect_to_previous(table):
+    page_ordering = session.get('%s.last_viewed' % table)
+    if not page_ordering:
+        return redirect(url_for('table_content', table=table))
+    page, ordering = page_ordering
+    kw = {}
+    if page and page != 1:
+        kw['page'] = page
+    if ordering:
+        kw['ordering'] = ordering
+    return redirect(url_for('table_content', table=table, **kw))
+
+@app.route('/<table>/update/<b64:pk>/', methods=['GET', 'POST'])
+@require_table
+def table_update(table, pk):
+    dataset.update_cache(table)
+    model = dataset[table].model_class
+    table_pk = model._meta.primary_key
+    if not table_pk:
+        flash('Table must have a primary key to perform update.', 'danger')
+        return redirect(url_for('table_content', table=table))
+    elif pk == '__uneditable__':
+        flash('Could not encode primary key to perform update.', 'danger')
+        return redirect(url_for('table_content', table=table))
+
+    expr = decode_pk(model, pk)
+    try:
+        obj = model.get(expr)
+    except model.DoesNotExist:
+        pk_repr = pk_display(table_pk, pk)
+        flash('Could not fetch row with primary-key %s.' % str(pk_repr), 'danger')
+        return redirect(url_for('table_content', table=table))
+
+    columns = dataset.get_columns(table)
+    col_dict = {}
+    row = {}
+    for column in columns:
+        value = getattr(obj, column.name)
+        if value is None:
+            row[column.name] = None
+        elif column.data_type.lower() == 'blob':
+            row[column.name] = base64.b64encode(value).decode('utf8')
+        else:
+            row[column.name] = value
+
+        col_dict[column.name] = column
+
+    edited = set()
+    errors = {}
+    if request.method == 'POST':
+        update = {}
+        for key, value in request.form.items():
+            if key not in col_dict: continue
+            column = col_dict[key]
+            edited.add(column.name)
+            row[column.name] = value
+
+            field = model._meta.columns[column.name]
+            value, err = minimal_validate_field(field, value)
+            if err:
+                errors[key] = err
+            else:
+                update[field] = value
+
+        if errors:
+            flash('One or more errors prevented the row being updated.',
+                  'danger')
+        elif update:
+            try:
+                with dataset.transaction() as txn:
+                    n = model.update(update).where(expr).execute()
+            except Exception as exc:
+                flash('Update failed: %s' % exc, 'danger')
+                app.logger.exception('Error attempting to update row from %s.', table)
+            else:
+                flash('Successfully updated %s record.' % n, 'success')
+                return redirect_to_previous(table)
+        else:
+            flash('No data was specified to be updated.', 'warning')
+
+    return render_template(
+        'table_update.html',
+        columns=columns,
+        edited=edited,
+        errors=errors,
+        model=model,
+        pk=pk,
+        row=row,
+        table=table,
+        table_pk=model._meta.primary_key)
+
+@app.route('/<table>/delete/<b64:pk>/', methods=['GET', 'POST'])
+@require_table
+def table_delete(table, pk):
+    dataset.update_cache(table)
+    model = dataset[table].model_class
+    table_pk = model._meta.primary_key
+    if not table_pk:
+        flash('Table must have a primary key to perform delete.', 'danger')
+        return redirect(url_for('table_content', table=table))
+    elif pk == '__uneditable__':
+        flash('Could not encode primary key to perform delete.', 'danger')
+        return redirect(url_for('table_content', table=table))
+
+    expr = decode_pk(model, pk)
+    try:
+        row = model.select().where(expr).dicts().get()
+    except model.DoesNotExist:
+        pk_repr = pk_display(table_pk, pk)
+        flash('Could not fetch row with primary-key %s.' % str(pk_repr), 'danger')
+        return redirect(url_for('table_content', table=table))
+
+    if request.method == 'POST':
+        try:
+            with dataset.transaction() as txn:
+                n = model.delete().where(expr).execute()
+        except Exception as exc:
+            flash('Delete failed: %s' % exc, 'danger')
+            app.logger.exception('Error attempting to delete row from %s.', table)
+        else:
+            flash('Successfully deleted %s record.' % n, 'success')
+            return redirect_to_previous(table)
+
+    return render_template(
+        'table_delete.html',
+        column_names=[c.name for c in dataset.get_columns(table)],
+        model=model,
+        pk=pk,
+        row=row,
+        table=table,
+        table_pk=table_pk)
+
 @app.route('/<table>/query/', methods=['GET', 'POST'])
 @require_table
 def table_query(table):
-    data = []
-    data_description = error = row_count = sql = None
+    return _query_view('table_query.html', table)
 
-    if request.method == 'POST':
-        sql = request.form['sql']
-        if 'export_json' in request.form:
-            return export(table, sql, 'json')
-        elif 'export_csv' in request.form:
-            return export(table, sql, 'csv')
-
-        try:
-            cursor = dataset.query(sql)
-        except Exception as exc:
-            error = str(exc)
-        else:
-            data = cursor.fetchall()[:app.config['MAX_RESULT_SIZE']]
-            data_description = cursor.description
-            row_count = cursor.rowcount
-    else:
-        if request.args.get('sql'):
-            sql = request.args.get('sql')
-        else:
-            sql = 'SELECT *\nFROM "%s"' % (table)
-
-    table_sql = dataset.query(
-        'SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = ?',
-        [table, 'table']).fetchone()[0]
-
-    return render_template(
-        'table_query.html',
-        data=data,
-        data_description=data_description,
-        error=error,
-        query_images=get_query_images(),
-        row_count=row_count,
-        sql=sql,
-        table=table,
-        table_sql=table_sql)
-
-@app.route('/table-definition/', methods=['POST'])
-def set_table_definition_preference():
-    key = 'show'
-    show = False
-    if request.form.get(key) and request.form.get(key) != 'false':
-        session[key] = show = True
-    elif key in session:
-        del session[key]
-    return jsonify({key: show})
-
-def export(table, sql, export_format):
-    model_class = dataset[table].model_class
-    query = model_class.raw(sql).dicts()
+def export(query, export_format, table=None):
     buf = StringIO()
     if export_format == 'json':
         kwargs = {'indent': 2}
-        filename = '%s-export.json' % table
+        filename = 'export.json'
         mimetype = 'text/javascript'
     else:
         kwargs = {}
-        filename = '%s-export.csv' % table
+        filename = 'export.csv'
         mimetype = 'text/csv'
 
+    if table:
+        filename = '%s-%s' % (table, filename)
+
+    # Avoid any special chars in export filename.
+    filename = re.sub(r'[^\w\d\-\.]+', '', filename)
+
     dataset.freeze(query, export_format, file_obj=buf, **kwargs)
 
     response_data = buf.getvalue()
     response = make_response(response_data)
     response.headers['Content-Length'] = len(response_data)
     response.headers['Content-Type'] = mimetype
-    response.headers['Content-Disposition'] = 'attachment; filename=%s' % (
+    response.headers['Content-Disposition'] = 'attachment; filename="%s"' % (
         filename)
     response.headers['Expires'] = 0
     response.headers['Pragma'] = 'public'
     return response
 
+@app.route('/<table>/export/', methods=['GET', 'POST'])
+@require_table
+def table_export(table):
+    columns = dataset.get_columns(table)
+    if request.method == 'POST':
+        export_format = request.form.get('export_format') or 'json'
+        col_dict = {c.name: c for c in columns}
+        selected = [c for c in (request.form.getlist('columns') or [])
+                    if c in col_dict]
+        if not selected:
+            flash('Please select one or more columns to export.', 'danger')
+        else:
+            model = dataset[table].model_class
+            fields = [model._meta.columns[c] for c in selected]
+            query = model.select(*fields).dicts()
+            try:
+                return export(query, export_format, table)
+            except Exception as exc:
+                flash('Error generating export: %s' % exc, 'danger')
+                app.logger.exception('Error generating export.')
+
+    return render_template(
+        'table_export.html',
+        columns=columns,
+        table=table)
+
 @app.route('/<table>/import/', methods=['GET', 'POST'])
 @require_table
 def table_import(table):
@@ -594,6 +967,7 @@ def table_import(table):
                         strict=strict)
             except Exception as exc:
                 flash('Error importing file: %s' % exc, 'danger')
+                app.logger.exception('Error importing file.')
             else:
                 flash(
                     'Successfully imported %s objects from %s.' % (
@@ -610,14 +984,26 @@ def table_import(table):
 @app.route('/<table>/drop/', methods=['GET', 'POST'])
 @require_table
 def drop_table(table):
+    is_view = any(v.name == table for v in dataset.get_all_views())
+    label = 'view' if is_view else 'table'
     if request.method == 'POST':
-        model_class = dataset[table].model_class
-        model_class.drop_table()
-        dataset.update_cache()  # Update all tables.
-        flash('Table "%s" dropped successfully.' % table, 'success')
-        return redirect(url_for('index'))
+        try:
+            if is_view:
+                dataset.query('DROP VIEW "%s";' % table)
+            else:
+                model_class = dataset[table].model_class
+                model_class.drop_table()
+        except Exception as exc:
+            flash('Error attempting to drop %s "%s".' % (label, table), 'danger')
+            app.logger.exception('Error attempting to drop %s "%s".', label, table)
+        else:
+            dataset.update_cache()  # Update all tables.
+            flash('%s "%s" dropped successfully.' %
+                  ('view' if is_view else 'table', table),
+                  'success')
+            return redirect(url_for('index'))
 
-    return render_template('drop_table.html', table=table)
+    return render_template('drop_table.html', is_view=is_view, table=table)
 
 @app.template_filter('format_index')
 def format_index(index_sql):
@@ -628,6 +1014,30 @@ def format_index(index_sql):
     create, definition = split_regex.split(index_sql)
     return '\nON '.join((create.strip(), definition.strip()))
 
+@app.template_filter('encode_pk')
+def encode_pk(row, pk):
+    if isinstance(pk, CompositeKey):
+        try:
+            return ':::'.join([str(row[k]) for k in pk.field_names])
+        except Exception as exc:
+            return '__uneditable__'
+    return row[pk.column_name]
+
+def decode_pk(model, pk_data):
+    pk = model._meta.primary_key
+    if isinstance(pk, CompositeKey):
+        fields = [pk.model._meta.columns[f] for f in pk.field_names]
+        values = pk_data.split(':::')
+        expressions = [(f == v) for f, v in zip(fields, values)]
+        return reduce(operator.and_, expressions)
+    return (pk == pk_data)
+
+@app.template_filter('pk_display')
+def pk_display(table_pk, pk):
+    if isinstance(table_pk, CompositeKey):
+        return tuple(pk.split(':::'))
+    return pk
+
 @app.template_filter('value_filter')
 def value_filter(value, max_length=50):
     if isinstance(value, numeric):
@@ -636,7 +1046,7 @@ def value_filter(value, max_length=50):
     if isinstance(value, binary_types):
         if not isinstance(value, (bytes, bytearray)):
             value = bytes(value)  # Handle `buffer` type.
-        value = value.decode('utf-8', decode_handler)
+        value = base64.b64encode(value)[:1024].decode('utf8')
     if isinstance(value, unicode_type):
         value = escape(value)
         if len(value) > max_length:
@@ -750,6 +1160,11 @@ def get_option_parser():
         default=True,
         dest='browser',
         help='Do not automatically open browser page.')
+    parser.add_option(
+        '-l',
+        '--log-file',
+        dest='log_file',
+        help='Filename for application logs.')
     parser.add_option(
         '-P',
         '--password',
@@ -762,6 +1177,13 @@ def get_option_parser():
         action='store_true',
         dest='read_only',
         help='Open database in read-only mode.')
+    parser.add_option(
+        '-R',
+        '--rows-per-page',
+        default=50,
+        dest='rows_per_page',
+        help='Number of rows to display per page (default=50)',
+        type='int')
     parser.add_option(
         '-u',
         '--url-prefix',
@@ -828,6 +1250,10 @@ def initialize_app(filename, read_only=False, password=None, url_prefix=None,
     if password:
         install_auth_handler(password)
 
+    dataset_kw = {}
+    if peewee_version >= (3, 14, 9):
+        dataset_kw['include_views'] = True
+
     if read_only:
         if sys.version_info < (3, 4, 0):
             die('Python 3.4.0 or newer is required for read-only access.')
@@ -840,17 +1266,19 @@ def initialize_app(filename, read_only=False, password=None, url_prefix=None,
             die('Unable to open database file in read-only mode. Ensure that '
                 'the database exists in order to use read-only mode.')
         db.close()
-        dataset = SqliteDataSet(db, bare_fields=True)
     else:
-        dataset = SqliteDataSet('sqlite:///%s' % filename, bare_fields=True)
+        db = SqliteDatabase(filename)
+
+    if extensions:
+        # Load extensions before performing introspection.
+        for ext in extensions:
+            db.load_extension(ext)
+
+    dataset = SqliteDataSet(db, bare_fields=True, **dataset_kw)
 
     if url_prefix:
         app.wsgi_app = PrefixMiddleware(app.wsgi_app, prefix=url_prefix)
 
-    if extensions:
-        for ext in extensions:
-            dataset._database.load_extension(ext)
-
     migrator = dataset._migrator
     dataset.close()
 
@@ -861,6 +1289,13 @@ def main():
     if not args:
         die('Error: missing required path to database file.')
 
+    if options.log_file:
+        fmt = logging.Formatter('[%(asctime)s] - [%(levelname)s] - %(message)s')
+        handler = WatchedFileHandler(options.log_file)
+        handler.setLevel(logging.DEBUG if options.debug else logging.WARNING)
+        handler.setFormatter(fmt)
+        app.logger.addHandler(handler)
+
     password = None
     if options.prompt_password:
         if os.environ.get('SQLITE_WEB_PASSWORD'):
@@ -874,6 +1309,9 @@ def main():
                 else:
                     break
 
+    if options.rows_per_page:
+        app.config['ROWS_PER_PAGE'] = options.rows_per_page
+
     # Initialize the dataset instance and (optionally) authentication handler.
     initialize_app(args[0], options.read_only, password, options.url_prefix,
                    options.extensions)
